@@ -167,7 +167,7 @@ function isString(obj) {
  * \return true if the given object is an RVector object.
  */
 function isVector(obj) {
-    return (typeof(obj)==="object" && obj.toString().startsWith("RVector"));
+    return (typeof(obj)==="object" && isOfType(obj, RVector));
 }
 
 /**
@@ -625,6 +625,15 @@ function isLeaderEntity(obj) {
     return isOfType(obj, RLeaderEntity) || isOfType(obj, RLeaderEntityPointer);
 }
 
+/**
+ * Checks if the given object is a tolerance entity.
+ *
+ * \return true if the given object is a tolerance entity (RToleranceEntity).
+ */
+function isToleranceEntity(obj) {
+    return isOfType(obj, RToleranceEntity) || isOfType(obj, RToleranceEntityPointer);
+}
+
 /**
  * Checks if the given object is an entity of the given type.
  */
@@ -863,9 +872,9 @@ function entityTypeToString(type, plural) {
     case RS.EntityDimAngular:
         return plural ? qsTr("Angular Dimensions") : qsTr("Angular Dimension");
     case RS.EntityDimAngular2L:
-        return plural ? qsTr("2 Line Angular Dimensions") : qsTr("2 Line Angular Dimension");
+        return plural ? qsTr("Angular Dimensions (2 Line)") : qsTr("Angular Dimension (2 Line)");
     case RS.EntityDimAngular3P:
-        return plural ? qsTr("3 Point Angular Dimensions") : qsTr("3 Point Angular Dimension");
+        return plural ? qsTr("Angular Dimensions (3 Point)") : qsTr("Angular Dimension (3 Point)");
     case RS.EntityDimArcLength:
         return plural ? qsTr("Arc Dimension") : qsTr("Arc Dimension");
     case RS.EntityDimDiametric:
@@ -884,6 +893,8 @@ function entityTypeToString(type, plural) {
         return plural ? qsTr("Images") : qsTr("Image");
     case RS.EntityLeader:
         return plural ? qsTr("Leaders") : qsTr("Leader");
+    case RS.EntityTolerance:
+        return plural ? qsTr("Tolerances") : qsTr("Tolerance");
     case RS.EntityLine:
         return plural ? qsTr("Lines") : qsTr("Line");
     case RS.EntityXLine:
@@ -914,6 +925,19 @@ function entityTypeToString(type, plural) {
     }
 }
 
+function translateFilterStrings(filterStrings) {
+    for (var i=0; i<filterStrings.length; i++) {
+        filterStrings[i] = filterStrings[i].replace("All CAD Files", qsTr("All CAD Files"));
+        filterStrings[i] = filterStrings[i].replace("DXF Files", qsTr("DXF Files"));
+        filterStrings[i] = filterStrings[i].replace("SVG Files", qsTr("SVG Files"));
+        filterStrings[i] = filterStrings[i].replace("DXF Drawing", qsTr("DXF Drawing"));
+        filterStrings[i] = filterStrings[i].replace("DWG Drawing", qsTr("DWG Drawing"));
+        filterStrings[i] = filterStrings[i].replace("PDF File", qsTr("PDF File"));
+        filterStrings[i] = filterStrings[i].replace("PDF/A-1B File", qsTr("PDF/A-1B File"));
+    }
+    return filterStrings;
+}
+
 /**
  * Checks the type of the given object.
  *
@@ -1339,6 +1363,21 @@ Array.prototype.remove = function(val, compareFunction) {
     }
 };
 
+/**
+ * Removes all matches of the given value from the array.
+ * \param The value(s) to remove.
+ */
+Array.prototype.removeAll = function() {
+    var what, a = arguments, l = a.length, ax;
+    while (l && this.length) {
+        what = a[--l];
+        while ((ax = this.indexOf(what)) !== -1) {
+            this.splice(ax, 1);
+        }
+    }
+    return this;
+};
+
 /**
  * Remove the first case insensitive match of the given string value from the array.
  * \param val The value to remove.
@@ -1384,63 +1423,8 @@ Array.prototype.sortNumerical = function() {
     Array.prototype.sort.call(this, function(a,b) { return a - b });
 }
 
-/* alphanum.js (C) Brian Huisman
- * Based on the Alphanum Algorithm by David Koelle
- * The Alphanum Algorithm is discussed at http://www.DaveKoelle.com
- *
- * Distributed under same license as original
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
- */
 Array.alphaNumericalSorter = function(a, b) {
-    if (RS.getSystemId()==="linux") {
-        // workaround for QCollator / Linux bug:
-        function chunkify(t) {
-            var tz = [];
-            var x = 0, y = -1, n = 0, i, j;
-
-            while (i = (j = t.charAt(x++)).charCodeAt(0)) {
-                var m = (i == 46 || (i >=48 && i <= 57));
-                if (m !== n) {
-                    tz[++y] = "";
-                    n = m;
-                }
-                tz[y] += j;
-            }
-            return tz;
-        }
-
-        var aa = chunkify(a.toLowerCase());
-        var bb = chunkify(b.toLowerCase());
-
-        for (x = 0; aa[x] && bb[x]; x++) {
-            if (aa[x] !== bb[x]) {
-                var c = Number(aa[x]), d = Number(bb[x]);
-                if (c == aa[x] && d == bb[x]) {
-                    return c - d;
-                } else return (aa[x] > bb[x]) ? 1 : -1;
-            }
-        }
-
-        return aa.length - bb.length;
-    }
-    else {
-        // Windows / macOS support alphanumerical sorting and
-        // sorting with international characters as well:
-        return RS.compareAlphanumerical(a, b);
-    }
+    return RS.compareAlphanumerical(a, b);
 };
 
 /**
@@ -1778,7 +1762,7 @@ function coordinateToString(coordinate, decimals, relative, polar, doc) {
     var second;
     var sep;
     if (polar) {
-        sep = RSettings.getStringValue("Input/PolarCoordinateSeparator", "<");
+        sep = RSettings.getPolarCoordinateSeparator();
         if (!isNull(doc)) {
             first = doc.formatLinear(coordinate.getMagnitude());
             second = doc.formatAngle(coordinate.getAngle());
@@ -1789,7 +1773,7 @@ function coordinateToString(coordinate, decimals, relative, polar, doc) {
         }
     }
     else {
-        sep = RSettings.getStringValue("Input/CartesianCoordinateSeparator", ",");
+        sep = RSettings.getCartesianCoordinateSeparator();
         if (!isNull(doc)) {
             first = doc.formatLinear(coordinate.x);
             second = doc.formatLinear(coordinate.y);
@@ -1802,10 +1786,7 @@ function coordinateToString(coordinate, decimals, relative, polar, doc) {
 
     var prefix;
     if (relative) {
-        prefix = RSettings.getStringValue(
-                    "Input/RelativeCoordinatePrefix",
-                    String.fromCharCode(64)  // @ (doxygen can't cope with an @ here)
-        );
+        prefix = RSettings.getRelativeCoordinatePrefix();
     }
     else {
         prefix = "";
@@ -1875,38 +1856,43 @@ function stringToDirectDistanceEntry(relativeZero, cursorPosition, str) {
  * Creates and returns a new entity based on the given shape.
  */
 function shapeToEntity(document, shape) {
-    if (isPointShape(shape)) {
-        return new RPointEntity(document, new RPointData(shape.getPosition()));
-    }
-    else if (isLineShape(shape)) {
-        return new RLineEntity(document, new RLineData(shape));
-    }
-    else if (isRayShape(shape)) {
-        return new RRayEntity(document, new RRayData(shape));
-    }
-    else if (isXLineShape(shape)) {
-        return new RXLineEntity(document, new RXLineData(shape));
-    }
-    else if (isArcShape(shape)) {
-        return new RArcEntity(document, new RArcData(shape));
-    }
-    else if (isCircleShape(shape)) {
-        return new RCircleEntity(document, new RCircleData(shape));
-    }
-    else if (isEllipseShape(shape)) {
-        return new REllipseEntity(document, new REllipseData(shape));
-    }
-    else if (isPolylineShape(shape)) {
-        return new RPolylineEntity(document, new RPolylineData(shape));
-    }
-    else if (isSplineShape(shape)) {
-        return new RSplineEntity(document, new RSplineData(shape));
-    }
-    else if (isTriangleShape(shape)) {
-        return new RSolidEntity(document, new RSolidData(shape));
+    var s = shape;
+    if (isFunction(s.data)) {
+        s = s.data();
     }
 
-    qWarning("shapeToEntity: unknown shape: ", shape);
+    if (isPointShape(s)) {
+        return new RPointEntity(document, new RPointData(s.getPosition()));
+    }
+    else if (isLineShape(s)) {
+        return new RLineEntity(document, new RLineData(s));
+    }
+    else if (isRayShape(s)) {
+        return new RRayEntity(document, new RRayData(s));
+    }
+    else if (isXLineShape(s)) {
+        return new RXLineEntity(document, new RXLineData(s));
+    }
+    else if (isArcShape(s)) {
+        return new RArcEntity(document, new RArcData(s));
+    }
+    else if (isCircleShape(s)) {
+        return new RCircleEntity(document, new RCircleData(s));
+    }
+    else if (isEllipseShape(s)) {
+        return new REllipseEntity(document, new REllipseData(s));
+    }
+    else if (isPolylineShape(s)) {
+        return new RPolylineEntity(document, new RPolylineData(s));
+    }
+    else if (isSplineShape(s)) {
+        return new RSplineEntity(document, new RSplineData(s));
+    }
+    else if (isTriangleShape(s)) {
+        return new RSolidEntity(document, new RSolidData(s));
+    }
+
+    qWarning("shapeToEntity: unknown shape: ", s);
     return undefined;
 }
 
@@ -2418,24 +2404,44 @@ function autoIconPath(path) {
 function applyTheme() {
     var theme = RSettings.getValue("Theme/ThemeName", undefined);
     if (!isNull(theme)) {
-        var prefix = "themes/" + theme + "/";
-        var fn = prefix + "stylesheet.css";
-        if (new QFileInfo(fn).exists()) {
-            var file = new QFile(fn);
-            var flags = new QIODevice.OpenMode(QIODevice.ReadOnly | QIODevice.Text);
-            if (file.open(flags)) {
-                var textStream = new QTextStream(file);
-                var allLines = textStream.readAll();
-                file.close();
-                allLines = allLines.replace(/url\(/g, "url(" + prefix);
-                qApp.setStyleSheet(allLines);
-                return;
+        var path = "themes/" + theme + "/";
+
+        qApp.styleSheet = "";
+
+        // load stylesheet.css, stylesheet_[win|macos|linux].css:
+        var found = false;
+        var systemId = RS.getSystemId();
+        if (systemId==="osx") systemId = "macos";
+
+        var postfixes = ["", "_" + systemId];
+        for (var i=0; i<postfixes.length; i++) {
+            var postfix = postfixes[i];
+            var fn = path + "stylesheet" + postfix + ".css";
+            qDebug("trying to load theme stylesheet: ", fn);
+
+            if (new QFileInfo(fn).exists()) {
+                var css = readTextFile(fn);
+                if (css.contains("RequiresPlugin:true")) {
+                    // only load theme if plugin loaded:
+                    var pluginId = theme.toUpperCase() + "STYLE";
+                    if (!RPluginLoader.hasPlugin(pluginId)) {
+                        qWarning("Theme not loaded: ", theme);
+                        qWarning("Theme plugin not found:", pluginId);
+                        return;
+                    }
+                }
+
+                css = css.replace(/url\(/g, "url(" + path);
+                qApp.styleSheet = qApp.styleSheet + "\n" + css;
+                found = true;
             }
         }
-        if (theme!=="Default") {
-            qWarning("Cannot open theme: ", theme);
+
+        if (!found) {
+            if (theme!=="Default") {
+                qWarning("Cannot open theme: ", theme);
+            }
         }
-        qApp.setStyleSheet("");
     }
 }
 
@@ -2444,6 +2450,7 @@ function readTextFile(fileName) {
     var flags = new QIODevice.OpenMode(QIODevice.ReadOnly | QIODevice.Text);
     if (file.open(flags)) {
         var textStream = new QTextStream(file);
+        textStream.setCodec("UTF-8");
         var contents = textStream.readAll();
         file.close();
         return contents;
@@ -2457,23 +2464,54 @@ function writeTextFile(fileName, str) {
     var flags = new QIODevice.OpenMode(QIODevice.WriteOnly | QIODevice.Text);
     if (file.open(flags)) {
         var textStream = new QTextStream(file);
+        textStream.setCodec("UTF-8");
         textStream.writeString(str);
     }
     file.close();
-};
+}
+
+function getKeyboardModifiers(event) {
+    if (!isNull(event)) {
+        if (isFunction(event.getModifiers)) {
+            return event.getModifiers();
+        }
+        else {
+            return event.modifiers();
+        }
+    }
 
-function getKeyboardModifiers() {
     if (RSettings.isQt(5)) {
-        return QGuiApplication.keyboardModifiers();
+        return QGuiApplication.queryKeyboardModifiers();
     }
     else {
-        return QApplication.keyboardModifiers();
+        return QApplication.queryKeyboardModifiers();
     }
-};
+}
 
-function isShiftPressed() {
-    return getKeyboardModifiers().valueOf() & Qt.ShiftModifier.valueOf();
-};
+function isShiftPressed(event) {
+    return (getKeyboardModifiers(event) & Qt.ShiftModifier) > 0 ||
+           (getKeyboardModifiers() & Qt.ShiftModifier) > 0;
+}
+
+function isAltPressed(event) {
+    return (getKeyboardModifiers(event) & Qt.AltModifier) > 0 ||
+           (getKeyboardModifiers() & Qt.AltModifier) > 0;
+}
+
+function isControlPressed(event) {
+    return (getKeyboardModifiers(event) & Qt.ControlModifier) > 0 ||
+           (getKeyboardModifiers() & Qt.ControlModifier) > 0;
+}
+
+function createSpatialIndex() {
+    if (!isNull(global.SpatialIndexClass)) {
+        // use class defined as global variable "SpatialIndexClass":
+        // allows plugins to override the default spatial index class:
+        return new global[global.SpatialIndexClass]();
+    }
+    //return new RSpatialIndexSimple();
+    return new RSpatialIndexNavel();
+}
 
 // Qt 4 API workaround:
 function qsTranslate2(context, sourceText, disambiguation, n) {
